import json

import requests
from qcodes_contrib_drivers.drivers.ERAInstruments import ERASynthPlusPlus
from qibo.config import log

from .oscillator import LocalOscillator, LocalOscillatorSettings

__all__ = ["ERASynth"]

RECONNECTION_ATTEMPTS = 10
"""Number of times to attempt sending requests to the web server in case of
failure."""
TIMEOUT = 10
"""Timeout time for HTTP requests in seconds."""


class ERASynthEthernet:
    """ERA ethernet driver that follows the QCoDeS interface.

    Controls the instrument via HTTP requests to the instrument's web
    server.
    """

    def __init__(self, name, address):
        self.name = name
        self.address = address
        self.post("readAll", 1)
        self.post("readDiagnostic", 0)
        self.post("rfoutput", 0)

    @property
    def url(self):
        return f"http://{self.address}/"

    def post(self, name, value):
        """Post a value to the instrument's web server.

        Try to post multiple times, waiting for 0.1 seconds between each attempt.

        Args:
            name: str = The name of the value to post.
            value: str = The value to post.
        """
        value = str(value)
        for _ in range(RECONNECTION_ATTEMPTS):
            try:
                response = requests.post(self.url, data={name: value}, timeout=TIMEOUT)
                if response.status_code == 200:
                    return True
                break
            except (ConnectionError, TimeoutError, requests.exceptions.ReadTimeout):
                log.info("ERAsynth connection timed out, retrying...")
        raise ConnectionError(f"Unable to post {name}={value} to {self.name}")

    def get(self, name):
        """Get a value from the instrument's web server.

        Try to get multiple times, waiting for 0.1 seconds between each attempt.

        Args:
            name: str = The name of the value to get.
        """
        if name == "ref_osc_source":
            value = self.get("reference_int_ext")
            if value == 1:
                return "EXT"
            else:
                return "INT"

        for _ in range(RECONNECTION_ATTEMPTS):
            try:
                response = requests.post(
                    self.url, params={"readAll": 1}, timeout=TIMEOUT
                )
                if response.status_code == 200:
                    # reponse.text is a dictonary in string format, convert it to a dictonary
                    return json.loads(response.text)[name]
                break
            except (ConnectionError, TimeoutError, requests.exceptions.ReadTimeout):
                log.info("ERAsynth connection timed out, retrying...")

        raise ConnectionError(f"Unable to get {name} from {self.name}")

    def set(self, name, value):
        """Set a value to the instrument's web server.

        Args:
            name (str): Name of the paramater that we are updating.
                In qibolab this can be ``frequency``, ``power`` or ``ref_osc_source``,
                however the instrument's web server may support more values.
            value: New value to set to the given parameter.
                The type of value depends on the parameter being updated.
        """
        if name == "ref_osc_source":
            if value.lower() in ("int", "internal"):
                self.post("reference_int_ext", 0)
            elif value.lower() in ("ext", "external"):
                self.post("reference_int_ext", 1)
            else:
                raise ValueError(f"Invalid reference clock source {value}")

        elif name == "frequency":
            self.post(name, int(value))

        elif name == "power":
            self.post(name, float(value))

        else:
            self.post(name, value)

    def on(self):
        self.post("rfoutput", 1)

    def off(self):
        self.post("rfoutput", 0)

    def close(self):
        self.off()


class ERASynth(LocalOscillator):
    """Driver to control the ERAsynth++ local oscillator.

    This driver is using:
    https://qcodes.github.io/Qcodes_contrib_drivers/api/generated/qcodes_contrib_drivers.drivers.ERAInstruments.html#qcodes_contrib_drivers.drivers.ERAInstruments.erasynth.ERASynthPlusPlus

    or the custom :class:`qibolab.instruments.erasynth.ERASynthEthernet` object
    if we are connected via ethernet.
    """

    def __init__(self, address, ethernet=True, ref_osc_source=None):
        super().__init__(
            address=address,
            settings=LocalOscillatorSettings(ref_osc_source=ref_osc_source),
        )
        self.ethernet = ethernet

    def create(self):
        name = f"{type(self).__name__}{id(self)}"
        if self.ethernet:
            return ERASynthEthernet(name, self.address)
        else:
            return ERASynthPlusPlus(name, f"TCPIP::{self.address}::INSTR")



import socket

import yaml
from pydantic import Field
from qibo.config import log

from qibolab._core.instruments.abstract import Instrument

__all__ = ["TemperatureController"]


class TemperatureController(Instrument):
    """Bluefors temperature controller.

    Example usage::

        if __name__ == "__main__":
            tc = TemperatureController("XLD1000_Temperature_Controller", "192.168.0.114", 8888)
            tc.connect()
            temperature_values = tc.read_data()
            for temperature_value in temperature_values:
                print(temperature_value)
    """

    address: str
    """IP address of the board sending cryo temperature data."""
    port: int = 8888
    """Port of the board sending cryo temperature data."""
    client_socket: socket.socket = Field(
        default_factory=lambda: socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    )
    _is_connected: bool = False

    def connect(self):
        """Connect to the socket."""
        if self._is_connected:
            return
        log.info(f"Bluefors connection. IP: {self.address} Port: {self.port}")
        self.client_socket.connect((self.address, self.port))
        self._is_connected = True
        log.info("Bluefors Temperature Controller Connected")

    def disconnect(self):
        """Disconnect from the socket."""
        if self._is_connected:
            self.client_socket.close()
            self._is_connected = False

    def get_data(self) -> dict[str, dict[str, float]]:
        """Connect to the socket and get temperature data.

        The typical message looks like this::

            flange_name: {'temperature':12.345678, 'timestamp':1234567890.123456}

        ``timestamp`` can be converted to datetime using ``datetime.fromtimestamp``.

        Returns:
            message (dict[str, dict[str, float]]): socket message in this format:
                {"flange_name": {'temperature': <value(float)>, 'timestamp':<value(float)>}}
        """
        return yaml.safe_load(self.client_socket.recv(1024).decode())

    def read_data(self):
        """Continously read data from the temperature controller."""
        while True:
            yield self.get_data()



"""Component (a.k.a. logical component) is a concept that is part of the
qibolab interface exposed to its users.

Interacting with, and controlling a quantum computer means orchestrating
its control electronics/equipment. For a qibolab user quantum computer
is just a collection of control instruments. A component represents a
piece of equipment, functionality, or configuration that can be
individually addressed, configured, and referred to in any relevant
context. It can represent a single device, part of a bigger device, or a
collection of multiple devices. Qibolab defines a handful of specific
categories of components, and each platform definition shall have only
such components, independent of what specific electronics is used, how
it is used, etc. One basic requirement for all components is that they
have unique names, and in any relevant context can be referred to by
their name.
"""

from . import channels
from .channels import *
from .configs import *

__all__ = []
__all__ += channels.__all__



"""Plotting tools for pulses and related entities."""

from collections import defaultdict
from typing import Optional

import matplotlib.pyplot as plt
import numpy as np

from qibolab._core.sequence import PulseSequence

from .envelope import Waveform
from .modulation import modulate
from .pulse import Delay, Pulse, VirtualZ

SAMPLING_RATE = 1
"""Default sampling rate in gigasamples per second (GSps).

Used for generating waveform envelopes if the instruments do not provide
a different value.
"""


def waveform(wf: Waveform, filename=None):
    """Plot the waveform.

    Args:
        filename (str): a file path. If provided the plot is save to a file.
    """
    plt.figure(figsize=(14, 5), dpi=200)
    plt.plot(wf, c="C0", linestyle="dashed")
    plt.xlabel("Sample Number")
    plt.ylabel("Amplitude")
    plt.grid(visible=True, which="both", axis="both", color="#888888", linestyle="-")
    if filename:
        plt.savefig(filename)
    else:
        plt.show()
    plt.close()


def pulse(pulse_: Pulse, freq: Optional[float] = None, filename: Optional[str] = None):
    """Plot the pulse envelope and modulated waveforms.

    Args:
        freq: Carrier frequency used to plot modulated waveform. None if modulated plot is not needed.
        filename (str): a file path. If provided the plot is save to a file.
    """
    import matplotlib.pyplot as plt
    from matplotlib import gridspec

    waveform_i = pulse_.i(SAMPLING_RATE)
    waveform_q = pulse_.q(SAMPLING_RATE)

    num_samples = len(waveform_i)
    time = np.arange(num_samples) / SAMPLING_RATE
    _ = plt.figure(figsize=(14, 5), dpi=200)
    gs = gridspec.GridSpec(ncols=2, nrows=1, width_ratios=np.array([2, 1]))
    ax1 = plt.subplot(gs[0])
    ax1.plot(
        time,
        waveform_i,
        label="envelope i",
        c="C0",
        linestyle="dashed",
    )
    ax1.plot(
        time,
        waveform_q,
        label="envelope q",
        c="C1",
        linestyle="dashed",
    )

    envelope = pulse_.envelopes(SAMPLING_RATE)
    modulated = (
        modulate(np.array(envelope), freq, rate=SAMPLING_RATE)
        if freq is not None
        else None
    )

    if modulated is not None:
        ax1.plot(time, modulated[0], label="modulated i", c="C0")
        ax1.plot(time, modulated[1], label="modulated q", c="C1")

    ax1.plot(time, -waveform_i, c="silver", linestyle="dashed")
    ax1.set_xlabel("Time [ns]")
    ax1.set_ylabel("Amplitude")

    ax1.grid(visible=True, which="both", axis="both", color="#888888", linestyle="-")
    start = 0
    finish = float(pulse_.duration)
    ax1.axis((start, finish, -1.0, 1.0))
    ax1.legend()

    ax2 = plt.subplot(gs[1])
    ax2.plot(waveform_i, waveform_q, label="envelope", c="C2")
    if modulated is not None:
        ax2.plot(modulated[0], modulated[1], label="modulated", c="C3")
        ax2.plot(
            modulated[0][0],
            modulated[1][0],
            marker="o",
            markersize=5,
            label="start",
            c="lightcoral",
        )
        ax2.plot(
            modulated[0][-1],
            modulated[1][-1],
            marker="o",
            markersize=5,
            label="finish",
            c="darkred",
        )

    ax2.plot(
        np.cos(time * 2 * np.pi / pulse_.duration),
        np.sin(time * 2 * np.pi / pulse_.duration),
        c="silver",
        linestyle="dashed",
    )

    ax2.grid(visible=True, which="both", axis="both", color="#888888", linestyle="-")
    ax2.legend()
    # ax2.axis([ -1, 1, -1, 1])
    ax2.axis("equal")
    if filename:
        plt.savefig(filename)
    else:
        plt.show()
    plt.close()


def sequence(ps: PulseSequence, freq: dict[str, float], filename=None):
    """Plot the sequence of pulses.

    Args:
        freq: frequency per channel, used to plot modulated waveforms of corresponding pulses. If a channel is missing from this dict,
         only /un-modulated/ waveforms are plotted for that channel.
        filename (str): a file path. If provided the plot is save to a file.
    """
    if len(ps) > 0:
        import matplotlib.pyplot as plt
        from matplotlib import gridspec

        # compile ``Align`` to delays as it is not supported here
        ps = ps.align_to_delays()
        num_pulses = len(ps)
        _ = plt.figure(figsize=(14, 2 * num_pulses), dpi=200)
        gs = gridspec.GridSpec(ncols=1, nrows=num_pulses)
        vertical_lines = []
        starts = defaultdict(float)
        for ch, pulse in ps:
            if not isinstance(pulse, Delay):
                vertical_lines.append(starts[ch])
                vertical_lines.append(starts[ch] + pulse.duration)
            starts[ch] += pulse.duration

        n = -1
        for ch in ps.channels:
            n += 1
            ax = plt.subplot(gs[n])
            ax.axis((0.0, ps.duration, -1.0, 1.0))
            start = 0
            for pulse in ps.channel(ch):
                if isinstance(pulse, (Delay, VirtualZ)):
                    start += pulse.duration
                    continue

                envelope = pulse.envelopes(SAMPLING_RATE)
                num_samples = envelope[0].size
                time = start + np.arange(num_samples) / SAMPLING_RATE
                if ch in freq:
                    modulated = modulate(
                        np.array(envelope), freq[ch], rate=SAMPLING_RATE
                    )
                    ax.plot(time, modulated[1], c="lightgrey")
                    ax.plot(time, modulated[0], c=f"C{str(n)}")
                ax.plot(time, pulse.i(SAMPLING_RATE), c=f"C{str(n)}")
                ax.plot(time, -pulse.i(SAMPLING_RATE), c=f"C{str(n)}")
                # TODO: if they overlap use different shades
                ax.axhline(0, c="dimgrey")
                ax.set_ylabel(f"channel {ch}")
                for vl in vertical_lines:
                    ax.axvline(vl, c="slategrey", linestyle="--")
                ax.axis((0, ps.duration, -1, 1))
                ax.grid(
                    visible=True,
                    which="both",
                    axis="both",
                    color="#CCCCCC",
                    linestyle="-",
                )
                start += pulse.duration

        if filename:
            plt.savefig(filename)
        else:
            plt.show()
        plt.close()



"""Bluefors drivers.

https://bluefors.com/
"""

from qibolab._core.instruments import bluefors
from qibolab._core.instruments.bluefors import *  # noqa: F403

__all__ = []
__all__ += bluefors.__all__



"""Dummy drivers.

Define instruments mainly used for testing purposes.
"""

from qibolab._core.instruments import dummy
from qibolab._core.instruments.dummy import *  # noqa: F403

__all__ = []
__all__ += dummy.__all__



"""ERA drivers.

http://erainstruments.com/
"""

from qibolab._core.instruments import erasynth
from qibolab._core.instruments.erasynth import *  # noqa: F403

__all__ = []
__all__ += erasynth.__all__



import numpy as np

from .envelope import IqWaveform

__all__ = ["wrap_phase", "rotate", "modulate", "demodulate"]


def wrap_phase(phase: float):
    """Limit phase to [0, 2pi)."""
    return phase % (2 * np.pi)


def rotate(envelope: IqWaveform, phase: float) -> IqWaveform:
    """Rotate envelopes in the IQ-plane according to a phase."""
    wrapped_phase = wrap_phase(phase)
    cos = np.cos(wrapped_phase)
    sin = np.sin(wrapped_phase)
    mod = np.array([[cos, -sin], [sin, cos]])
    return np.einsum("ij,jt->it", mod, envelope)


def modulate(
    envelope: IqWaveform,
    freq: float,
    rate: float,
    phase: float = 0.0,
) -> IqWaveform:
    """Modulate the envelope waveform with a carrier.

    `envelope` is a `(2, n)`-shaped array of I and Q (first dimension) envelope signals,
    as a function of time (second dimension), and `freq` the frequency of the carrier to
    modulate with (usually the IF) in GHz.
    `rate` is an optional sampling rate, in Gs/s, to sample the carrier.

    .. note::

        Only the combination `freq / rate` is actually relevant, but it is frequently
        convenient to specify one in GHz and the other in Gs/s. Thus the two arguments
        are provided for the simplicity of their interpretation.

    `phase` is an optional initial phase for the carrier.
    """
    samples = np.arange(envelope.shape[1])
    phases = (2 * np.pi * freq / rate) * samples + phase
    cos = np.cos(phases)
    sin = np.sin(phases)
    mod = np.array([[cos, -sin], [sin, cos]])

    # the normalization is related to `mod`, but only applied at the end for the sake of
    # performances
    return np.einsum("ijt,jt->it", mod, envelope) / np.sqrt(2)


def demodulate(
    modulated: IqWaveform,
    freq: float,
    rate: float,
) -> IqWaveform:
    """Demodulate the acquired pulse.

    The role of the arguments is the same of the corresponding ones in :func:`modulate`,
    which is essentially the inverse of this function.
    """
    # in case the offsets have not been removed in hardware
    modulated = modulated - np.mean(modulated)

    samples = np.arange(modulated.shape[1])
    phases = (2 * np.pi * freq / rate) * samples
    cos = np.cos(phases)
    sin = np.sin(phases)
    demod = np.array([[cos, sin], [-sin, cos]])

    # the normalization is related to `demod`, but only applied at the end for the sake
    # of performances
    return np.sqrt(2) * np.einsum("ijt,jt->it", demod, modulated)



from dataclasses import dataclass, field
from typing import Any, Generic, TypeVar

from qibolab._core.components import OscillatorConfig

from ..components import OpxOutputConfig, QmAcquisitionConfig

__all__ = ["AnalogOutput", "OctaveOutput", "OctaveInput", "Controller", "Octave"]


DEFAULT_INPUTS = {"1": {}, "2": {}}
"""Default controller config section.

Inputs are always registered to avoid issues with automatic mixer
calibration when using Octaves.
"""

V = TypeVar("V")


class PortDict(Generic[V], dict[str, V]):
    """Dictionary that automatically converts keys to strings.

    Used to register input and output ports to controllers and Octaves
    in the QUA config.
    """

    def __setitem__(self, key: Any, value: V):
        super().__setitem__(str(key), value)


@dataclass(frozen=True)
class AnalogOutput:
    offset: float = 0.0
    filter: dict[str, float] = field(default_factory=dict)

    @classmethod
    def from_config(cls, config: OpxOutputConfig):
        return cls(offset=config.offset, filter=config.filter)


@dataclass(frozen=True)
class AnalogInput:
    offset: float = 0.0
    gain_db: int = 0

    @classmethod
    def from_config(cls, config: QmAcquisitionConfig):
        return cls(offset=config.offset, gain_db=config.gain)


@dataclass(frozen=True)
class OctaveOutput:
    LO_frequency: int
    gain: int = 0
    LO_source: str = "internal"
    output_mode: str = "triggered"

    @classmethod
    def from_config(cls, config: OscillatorConfig):
        return cls(LO_frequency=config.frequency, gain=config.power)


@dataclass(frozen=True)
class OctaveInput:
    LO_frequency: int
    LO_source: str = "internal"
    IF_mode_I: str = "direct"
    IF_mode_Q: str = "direct"


@dataclass
class Controller:
    analog_outputs: PortDict[dict[str, AnalogOutput]] = field(default_factory=PortDict)
    digital_outputs: PortDict[dict[str, dict]] = field(default_factory=PortDict)
    analog_inputs: PortDict[dict[str, AnalogInput]] = field(
        default_factory=lambda: PortDict(DEFAULT_INPUTS)
    )

    def add_octave_output(self, port: int):
        # TODO: Add offset here?
        self.analog_outputs[2 * port - 1] = AnalogOutput()
        self.analog_outputs[2 * port] = AnalogOutput()

        self.digital_outputs[2 * port - 1] = {}

    def add_octave_input(self, port: int, config: QmAcquisitionConfig):
        self.analog_inputs[2 * port - 1] = self.analog_inputs[2 * port] = (
            AnalogInput.from_config(config)
        )


@dataclass
class Octave:
    connectivity: str
    RF_outputs: PortDict[dict[str, OctaveOutput]] = field(default_factory=PortDict)
    RF_inputs: PortDict[dict[str, OctaveInput]] = field(default_factory=PortDict)



import qcodes.instrument_drivers.rohde_schwarz.SGS100A as LO_SGS100A

from qibolab._core.instruments.oscillator import LocalOscillator

__all__ = ["SGS100A"]


class SGS100A(LocalOscillator):
    """Driver to control the Rohde-Schwarz SGS100A local oscillator.

    This driver is using:
    https://qcodes.github.io/Qcodes/api/generated/qcodes.instrument_drivers.rohde_schwarz.html#module-qcodes.instrument_drivers.rohde_schwarz.SGS100A
    """

    def create(self):
        name = f"{type(self).__name__}{id(self)}"
        return LO_SGS100A.RohdeSchwarz_SGS100A(
            name, f"TCPIP0::{self.address}::5025::SOCKET"
        )



from typing import Annotated, Optional

from pydantic import ConfigDict, Field

from .identifier import ChannelId, QubitId, TransitionId
from .serialize import Model

__all__ = ["Qubit"]

DefaultChannelType = Annotated[Optional[ChannelId], True]
"""If ``True`` the channel is included in the default qubit constructor."""


class Qubit(Model):
    """Representation of a physical qubit.

    Contains the channel ids used to control the qubit and is instantiated
    in the function that creates the corresponding
    :class:`qibolab.Platform`
    """

    model_config = ConfigDict(frozen=False)

    drive: DefaultChannelType = None
    """Ouput channel, to drive the qubit state."""
    drive_qudits: Annotated[dict[TransitionId, ChannelId], False] = Field(
        default_factory=dict
    )
    """Output channels collection, to drive non-qubit transitions."""
    flux: DefaultChannelType = None
    """Output channel, to control the qubit flux."""
    probe: DefaultChannelType = None
    """Output channel, to probe the resonator."""
    acquisition: DefaultChannelType = None
    """Input channel, to acquire the readout results."""

    @property
    def channels(self) -> list[ChannelId]:
        return [
            x
            for x in (
                [getattr(self, ch) for ch in ["probe", "acquisition", "drive", "flux"]]
                + list(self.drive_qudits.values())
            )
            if x is not None
        ]

    @classmethod
    def default(cls, name: QubitId, channels: Optional[list[str]] = None, **kwargs):
        """Create a qubit with default channel names.

        Default channel names follow the convention:
        '{qubit_name}/{channel_type}'

        Args:
            name: Name for the qubit to be used for channel ids.
            channels: List of channels to add to the qubit.
                If ``None`` the following channels will be added:
                probe, acquisition, drive and flux.
        """
        if channels is None:
            channels = [name for name, f in cls.model_fields.items() if f.metadata[0]]
        return cls(**{ch: f"{name}/{ch}" for ch in channels}, **kwargs)


class QubitPair(Model):
    """Represent a two-qubit interaction."""

    drive: Optional[ChannelId] = None
    """Output channel, for cross-resonance driving."""



"""Rohde & Schwarz drivers.

https://www.rohde-schwarz.com/
"""

from qibolab._core.instruments import rohde_schwarz
from qibolab._core.instruments.rohde_schwarz import *  # noqa: F403

__all__ = []
__all__ += rohde_schwarz.__all__



from typing import Optional

from qm import qua
from qm.qua import declare, fixed, for_
from qualang_tools.loops import from_array

from qibolab._core.execution_parameters import AcquisitionType, ExecutionParameters
from qibolab._core.pulses import Align, Delay, Pulse, Readout, VirtualZ
from qibolab._core.sweeper import ParallelSweepers, Parameter, Sweeper

from ..config import operation
from .acquisition import Acquisition
from .arguments import ExecutionArguments, Parameters
from .sweepers import INT_TYPE, NORMALIZERS, SWEEPER_METHODS, normalize_phase


def _delay(pulse: Delay, element: str, parameters: Parameters):
    # TODO: How to play delays on multiple elements?
    if parameters.duration is None:
        duration = max(int(pulse.duration) // 4 + 1, 4)
        qua.wait(duration, element)
    elif parameters.interpolated:
        duration = parameters.duration + 1
        qua.wait(duration, element)
    else:
        duration = parameters.duration / 4
        with qua.if_(duration < 4):
            qua.wait(4, element)
        with qua.else_():
            qua.wait(duration, element)


def _play_multiple_waveforms(element: str, parameters: Parameters):
    """Sweeping pulse duration using distinctly uploaded waveforms."""
    with qua.switch_(parameters.duration, unsafe=True):
        for value, sweep_op in parameters.duration_ops:
            if parameters.amplitude is not None:
                sweep_op = sweep_op * parameters.amplitude
            with qua.case_(value):
                qua.play(sweep_op, element)


def _play_single_waveform(
    op: str,
    element: str,
    parameters: Parameters,
    acquisition: Optional[Acquisition] = None,
):
    if parameters.amplitude is not None:
        op = parameters.amplitude_op * parameters.amplitude
    if acquisition is not None:
        acquisition.measure(op)
    else:
        qua.play(op, element, duration=parameters.duration)


def _play(
    op: str,
    element: str,
    parameters: Parameters,
    acquisition: Optional[Acquisition] = None,
):
    if parameters.phase is not None:
        qua.frame_rotation_2pi(parameters.phase, element)

    if len(parameters.duration_ops) > 0:
        _play_multiple_waveforms(element, parameters)
    else:
        _play_single_waveform(op, element, parameters, acquisition)

    if parameters.phase is not None:
        qua.reset_frame(element)


def play(args: ExecutionArguments):
    """Part of QUA program that plays an arbitrary pulse sequence.

    Should be used inside a ``program()`` context.
    """
    qua.align()

    # keep track of ``Align`` command that were already played
    # because the same ``Align`` will appear on multiple channels
    # in the sequence
    processed_aligns = set()

    for channel_id, pulse in args.sequence:
        element = str(channel_id)
        op = operation(pulse)
        params = args.parameters[pulse.id]
        if isinstance(pulse, Delay):
            _delay(pulse, element, params)
        elif isinstance(pulse, Pulse):
            _play(op, element, params)
        elif isinstance(pulse, Readout):
            acquisition = args.acquisitions.get((op, element))
            _play(op, element, params, acquisition)
        elif isinstance(pulse, VirtualZ):
            qua.frame_rotation_2pi(normalize_phase(pulse.phase), element)
        elif isinstance(pulse, Align) and pulse.id not in processed_aligns:
            channel_ids = args.sequence.pulse_channels(pulse.id)
            qua.align(*(str(ch) for ch in channel_ids))
            processed_aligns.add(pulse.id)

    if args.relaxation_time > 0:
        qua.wait(args.relaxation_time // 4)


def _process_sweeper(sweeper: Sweeper, args: ExecutionArguments):
    parameter = sweeper.parameter
    if parameter not in SWEEPER_METHODS:
        raise NotImplementedError(f"Sweeper for {parameter} is not implemented.")

    if parameter in INT_TYPE:
        variable = declare(int)
        values = sweeper.values.astype(int)
    else:
        variable = declare(fixed)
        values = sweeper.values

    if parameter is Parameter.frequency:
        lo_frequency = args.parameters[sweeper.channels[0]].lo_frequency
        values = NORMALIZERS[parameter](values, lo_frequency)
    elif parameter in NORMALIZERS:
        values = NORMALIZERS[parameter](values)

    return variable, values


def sweep(
    sweepers: list[ParallelSweepers],
    args: ExecutionArguments,
):
    """Unrolls a list of qibolab sweepers to the corresponding QUA for loops.

    Uses recursion to handle nested sweepers.
    """
    if len(sweepers) > 0:
        parallel_sweepers = sweepers[0]

        variables, all_values = zip(
            *(_process_sweeper(sweeper, args) for sweeper in parallel_sweepers)
        )
        if len(parallel_sweepers) > 1:
            loop = qua.for_each_(variables, all_values)
        else:
            loop = for_(*from_array(variables[0], all_values[0]))

        with loop:
            for sweeper, variable, values in zip(
                parallel_sweepers, variables, all_values
            ):
                method = SWEEPER_METHODS[sweeper.parameter]
                if sweeper.pulses is not None:
                    for pulse in sweeper.pulses:
                        params = args.parameters[pulse.id]
                        method(variable, params)
                else:
                    for channel in sweeper.channels:
                        params = args.parameters[channel]
                        method(variable, params)

            sweep(sweepers[1:], args)

    else:
        play(args)


def program(
    args: ExecutionArguments,
    options: ExecutionParameters,
    sweepers: list[ParallelSweepers],
):
    """QUA program implementing the required experiment."""
    with qua.program() as experiment:
        n = declare(int)
        # declare acquisition variables
        for acquisition in args.acquisitions.values():
            acquisition.declare()
        # execute pulses
        with for_(n, 0, n < options.nshots, n + 1):
            sweep(list(sweepers), args)
        # download acquisitions
        has_iq = options.acquisition_type is AcquisitionType.INTEGRATION
        buffer_dims = options.results_shape(sweepers)[::-1][int(has_iq) :]
        with qua.stream_processing():
            for acquisition in args.acquisitions.values():
                acquisition.download(*buffer_dims)
    return experiment



from typing import Annotated, Union

import numpy as np
import numpy.typing as npt
from pydantic import BeforeValidator, Field, PlainSerializer

QubitId = Annotated[Union[int, str], Field(union_mode="left_to_right")]
"""Qubit name."""

QubitPairId = Annotated[
    tuple[QubitId, QubitId],
    BeforeValidator(lambda p: tuple(p.split("-")) if isinstance(p, str) else p),
    PlainSerializer(lambda p: f"{p[0]}-{p[1]}"),
]
"""Type for holding ``QubitPair``s in the ``platform.pairs`` dictionary."""


ChannelId = str
"""Unique identifier for a channel."""


StateId = int
"""State identifier."""


def _split(pair: Union[str, tuple]) -> tuple[str, str]:
    if isinstance(pair, str):
        a, b = pair.split("-")
        return a, b
    return pair


def _join(pair: tuple[str, str]) -> str:
    return f"{pair[0]}-{pair[1]}"


TransitionId = Annotated[
    tuple[StateId, StateId], BeforeValidator(_split), PlainSerializer(_join)
]
"""Identifier for a state transition."""

QubitPairId = Annotated[
    tuple[QubitId, QubitId], BeforeValidator(_split), PlainSerializer(_join)
]
"""Two-qubit active interaction identifier."""


Result = npt.NDArray[np.float64]
"""An array of results returned by instruments."""



import numpy as np
import numpy.typing as npt
from qm import qua
from qm.qua._dsl import _Variable  # for type declaration only

from qibolab._core.components import Channel, Config
from qibolab._core.identifier import ChannelId
from qibolab._core.sweeper import Parameter

from .arguments import ExecutionArguments, Parameters

MAX_OFFSET = 0.5
"""Maximum voltage supported by Quantum Machines OPX+ instrument in volts."""
MAX_AMPLITUDE_FACTOR = 1.99
"""Maximum multiplication factor for ``qua.amp`` used when sweeping amplitude.

https://docs.quantum-machines.co/1.2.0/docs/API_references/qua/dsl_main/#qm.qua._dsl.amp
"""
FREQUENCY_BANDWIDTH = 4e8
"""Quantum Machines OPX+ frequency bandwidth in Hz."""


def find_lo_frequencies(
    args: ExecutionArguments,
    channels: list[tuple[ChannelId, Channel]],
    configs: dict[str, Config],
    values: npt.NDArray,
):
    """Register LO frequencies of swept channels in execution arguments.

    These are needed to calculate the proper IF when sweeping frequency.
    It also checks if frequency sweep is within the supported instrument
    bandwidth [-400, 400] MHz.
    """
    lo_freqs = {configs[channel.lo].frequency for _, channel in channels}
    if len(lo_freqs) > 1:
        raise ValueError(
            "Cannot sweep frequency of channels using different LO using the same `Sweeper` object. Please use parallel sweepers instead."
        )
    lo_frequency = lo_freqs.pop()
    for id, channel in channels:
        max_freq = max(abs(values - lo_frequency))
        if max_freq > FREQUENCY_BANDWIDTH:
            raise ValueError(
                f"Frequency {max_freq} for channel {id} is beyond instrument bandwidth."
            )
        args.parameters[id].lo_frequency = int(lo_frequency)


def sweeper_amplitude(values: npt.NDArray) -> float:
    """Pulse amplitude to be registered in the QM ``config`` when sweeping
    amplitude.

    The multiplicative factor used in the ``qua.amp`` command is limited, so we
    may need to register a pulse with different amplitude than the original pulse
    in the sequence, in order to reach all sweeper values when sweeping amplitude.
    """
    return max(abs(values)) / MAX_AMPLITUDE_FACTOR


def normalize_amplitude(values: npt.NDArray) -> npt.NDArray:
    """Normalize amplitude factor to [-MAX_AMPLITUDE_FACTOR,
    MAX_AMPLITUDE_FACTOR]."""
    return values / sweeper_amplitude(values)


def normalize_phase(values: npt.NDArray) -> npt.NDArray:
    """Normalize phase from [0, 2pi] to [0, 1]."""
    return values / (2 * np.pi)


def normalize_duration(values: npt.NDArray) -> npt.NDArray:
    """Convert duration from ns to clock cycles (clock cycle = 4ns)."""
    if any(values < 16) or not all(values % 4 == 0):
        raise ValueError(
            "Cannot use interpolated duration sweeper for durations that are not multiple of 4ns or are less than 16ns. Please use normal duration sweeper."
        )
    return (values // 4).astype(int)


def normalize_frequency(values: npt.NDArray, lo_frequency: int) -> npt.NDArray:
    """Convert frequencies to integer and subtract LO frequency.

    QUA gives an error if the raw frequency values are uploaded to sweep
    over.
    """
    return (values - lo_frequency).astype(int)


def _amplitude(variable: _Variable, parameters: Parameters):
    parameters.amplitude = qua.amp(variable)


def _relative_phase(variable: _Variable, parameters: Parameters):
    parameters.phase = variable


def _duration(variable: _Variable, parameters: Parameters):
    parameters.duration = variable


def _duration_interpolated(variable: _Variable, parameters: Parameters):
    parameters.duration = variable
    parameters.interpolated = True


def _offset(variable: _Variable, parameters: Parameters):
    with qua.if_(variable >= MAX_OFFSET):
        qua.set_dc_offset(parameters.element, "single", MAX_OFFSET)
    with qua.elif_(variable <= -MAX_OFFSET):
        qua.set_dc_offset(parameters.element, "single", -MAX_OFFSET)
    with qua.else_():
        qua.set_dc_offset(parameters.element, "single", variable)


def _frequency(variable: _Variable, parameters: Parameters):
    qua.update_frequency(parameters.element, variable)


INT_TYPE = {Parameter.frequency, Parameter.duration, Parameter.duration_interpolated}
"""Sweeper parameters for which we need ``int`` variable type.

The rest parameters need ``fixed`` type.
"""

NORMALIZERS = {
    Parameter.frequency: normalize_frequency,
    Parameter.amplitude: normalize_amplitude,
    Parameter.relative_phase: normalize_phase,
    Parameter.duration_interpolated: normalize_duration,
}
"""Functions to normalize sweeper values.

The rest parameters do not need normalization (identity function).
"""

SWEEPER_METHODS = {
    Parameter.frequency: _frequency,
    Parameter.amplitude: _amplitude,
    Parameter.duration: _duration,
    Parameter.duration_interpolated: _duration_interpolated,
    Parameter.relative_phase: _relative_phase,
    Parameter.offset: _offset,
}
"""Methods that return part of QUA program to be used inside the loop."""



"""Pulse class."""

from typing import Annotated, Literal, Union

import numpy as np
from pydantic import Field

from ..serialize import Model
from .envelope import Envelope, IqWaveform, Waveform

__all__ = [
    "Acquisition",
    "Align",
    "Delay",
    "Pulse",
    "PulseId",
    "PulseLike",
    "Readout",
    "VirtualZ",
]

PulseId = int
"""Unique identifier for a pulse."""


class _PulseLike(Model):
    @property
    def id(self) -> PulseId:
        """Instruction identifier."""
        return id(self)


class Pulse(_PulseLike):
    """A pulse to be sent to the QPU.

    Valid on any channel, except acquisition ones.
    """

    kind: Literal["pulse"] = "pulse"

    duration: float
    """Pulse duration."""

    amplitude: float
    """Pulse digital amplitude (unitless).

    Pulse amplitudes are normalised between -1 and 1.
    """
    envelope: Envelope
    """The pulse envelope shape.

    See :class:`qibolab.Envelope` for list of available shapes.
    """
    relative_phase: float = 0.0
    """Relative phase of the pulse, in radians."""

    def i(self, sampling_rate: float) -> Waveform:
        """Compute the envelope of the waveform i component."""
        samples = int(self.duration * sampling_rate)
        return self.amplitude * self.envelope.i(samples)

    def q(self, sampling_rate: float) -> Waveform:
        """Compute the envelope of the waveform q component."""
        samples = int(self.duration * sampling_rate)
        return self.amplitude * self.envelope.q(samples)

    def envelopes(self, sampling_rate: float) -> IqWaveform:
        """Compute a tuple with the i and q envelopes."""
        return np.array([self.i(sampling_rate), self.q(sampling_rate)])


class Delay(_PulseLike):
    """Wait instruction.

    During its length no pulse is sent on the same channel.

    Valid on any channel.
    """

    kind: Literal["delay"] = "delay"

    duration: float
    """Duration in ns."""


class VirtualZ(_PulseLike):
    """Implementation of Z-rotations using virtual phase.

    Only valid on a drive channel.
    """

    kind: Literal["virtualz"] = "virtualz"

    phase: float
    """Phase that implements the rotation."""

    @property
    def duration(self):
        """Duration of the virtual gate should always be zero."""
        return 0


class Acquisition(_PulseLike):
    """Acquisition instruction.

    This event instructs the device to acquire samples for the event
    span.

    Only valid on an acquisition channel.
    """

    kind: Literal["acquisition"] = "acquisition"

    duration: float
    """Duration in ns."""


class Readout(_PulseLike):
    """Readout instruction.

    This event instructs the device to acquire samples for the event
    span.

    Only valid on an acquisition channel.
    """

    kind: Literal["readout"] = "readout"

    acquisition: Acquisition
    probe: Pulse

    @classmethod
    def from_probe(cls, probe: Pulse):
        """Create a whole readout operation from its probe pulse.

        The acquisition is made to match the same probe duration.
        """
        return cls(acquisition=Acquisition(duration=probe.duration), probe=probe)

    @property
    def duration(self) -> float:
        """Duration in ns."""
        return self.acquisition.duration

    @property
    def id(self) -> int:
        """Instruction identifier."""
        return self.acquisition.id


class Align(_PulseLike):
    """Brings different channels at the same point in time."""

    kind: Literal["align"] = "align"


PulseLike = Annotated[
    Union[Align, Pulse, Delay, VirtualZ, Acquisition, Readout],
    Field(discriminator="kind"),
]
